# Generated by Django 2.1.14 on 2020-03-27 10:45

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0055_auto_20200325_1335'),
        ('djautotask', '0055_auto_20200325_0928'),
    ]

    operations = [
    ]



# Generated by Django 4.0.7 on 2022-10-20 17:24

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0109_billingcode_billing_code_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='account',
            name='phone',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-20 16:17

from django.db import migrations, models
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0081_project_department'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectUDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=50, unique=True)),
                ('label', models.CharField(blank=True, max_length=50, null=True)),
                ('type', models.CharField(blank=True, max_length=50, null=True)),
                ('is_picklist', models.BooleanField(default=False)),
                ('picklist', models.JSONField(default=dict)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TaskUDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=50, unique=True)),
                ('label', models.CharField(blank=True, max_length=50, null=True)),
                ('type', models.CharField(blank=True, max_length=50, null=True)),
                ('is_picklist', models.BooleanField(default=False)),
                ('picklist', models.JSONField(default=dict)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TicketUDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=50, unique=True)),
                ('label', models.CharField(blank=True, max_length=50, null=True)),
                ('type', models.CharField(blank=True, max_length=50, null=True)),
                ('is_picklist', models.BooleanField(default=False)),
                ('picklist', models.JSONField(default=dict)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='ticket',
            name='udf',
            field=models.JSONField(blank=True, default=dict, null=True),
        ),
    ]



# Generated by Django 2.1.14 on 2020-08-07 15:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0070_auto_20200714_0807'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='status_detail',
            field=models.CharField(blank=True, max_length=2000, null=True),
        ),
    ]



# Generated by Django 2.1.11 on 2019-08-29 10:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='ticket_number',
            field=models.CharField(blank=True, max_length=50, null=True),
        ),
    ]



# Generated by Django 4.0.7 on 2022-08-17 20:48

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0105_tasknote_created_by_contact'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticketnote',
            name='created_by_contact',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djautotask.contact'),
        ),
    ]



import responses
import requests

from django.core.cache import cache
from django.test import TestCase
from django.utils import timezone

from . import mocks as mk

from .. import api
from ..api import AutotaskAPIError, AutotaskAPIClientError, \
    ApiConditionList
from ..api import ApiCondition as A


class TestApiConditionList(TestCase):

    def setUp(self):
        self.conditions = ApiConditionList()

    def test_build_query_string_single_get(self):
        c = A(op='eq', field='isActive', value='true')
        self.conditions.add(c)
        endpoint, filters = self.conditions.build_query()
        self.assertEqual(
            endpoint,
            'query?search={"filter": [{"op": "eq", "field": "isActive", '
            '"value": "true"}]}'
        )
        self.assertEqual(filters, None)

    def test_build_query_string_multiple_post(self):
        test_datetime = timezone.datetime(2019, 6, 22, 2, 0, 0,
                                          tzinfo=timezone.utc)

        c = A(
            A(op='eq', field='isActive', value='true'),
            A(
                field='lastActivityDate',
                value=test_datetime.strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
                op='gt'
            ),
            op='and'
        )

        self.conditions.add(c)
        endpoint, filters = self.conditions.build_query(method="post")
        str_built = '{"filter": [{"op": "and", "items": [{"op": "eq", ' \
                    '"field": "isActive", "value": "true"}, {"op": "gt", ' \
                    '"field": "lastActivityDate", "value": ' \
                    '"2019-06-22T02:00:00.000000Z"}]}]}'
        self.assertEqual(endpoint, 'query')
        self.assertEqual(filters, str_built)


class TestAutotaskAPIClient(TestCase):
    API_URL = 'https://localhost/'

    def setUp(self):
        mk.init_zone_info_connection(return_value={
            'url': self.API_URL,
            'webUrl': self.API_URL,
        })
        self.client = api.ContactsAPIClient()  # Must use a real client as
        # AutotaskAPIClient is effectively abstract

    @responses.activate
    def test_request(self):
        endpoint = 'http://example.com/'
        response = 'ok'
        responses.add(responses.GET, endpoint, json=response, status=200)

        result = self.client.request('get', endpoint, None)
        self.assertEqual(result, response)

    @responses.activate
    def test_request_failure(self):
        endpoint = 'http://example.com/'
        responses.add(
            responses.GET, endpoint, body=requests.RequestException()
        )

        with self.assertRaises(AutotaskAPIError):
            self.client.request('get', endpoint, None)

    @responses.activate
    def test_request_400(self):
        endpoint = 'http://example.com/'
        response = {'error': 'this is bad'}
        responses.add(responses.GET, endpoint, json=response, status=400)

        with self.assertRaises(AutotaskAPIError):
            self.client.request('get', endpoint, None)


class TestFetchAPIUrl(TestCase):
    API_URL = 'https://localhost/'

    def test_fetch_api_url_added_to_cache(self):
        """
        API url should be fetched to the cache after request,
        when zone info is not found in the cache
        """
        cache.clear()
        self.assertIsNone(api.get_cached_url('url'))
        mk.init_zone_info_connection(return_value={
            'url': self.API_URL,
            'webUrl': self.API_URL,
        })
        api.ContactsAPIClient()
        self.assertIsNotNone(api.get_cached_url('url'))

    def test_fetch_api_url_from_warm_cache(self):
        """
        Request shouldn't be tried, when zone info is found in the cache
        """
        # get api url from cache during __init__
        cache.set('zone_url', 'some api url')
        mk.init_zone_info_connection(return_value={
            'url': "some NEW api url",
            'webUrl': "some NEW api url",
        })
        api.ContactsAPIClient()
        self.assertEqual(api.get_cached_url('url'), 'some api url')

    def test_get_specific_api_connection_url(self):
        mk.init_zone_info_connection(return_value={
            'url': self.API_URL,
            'webUrl': self.API_URL,
        })
        client = api.ContactsAPIClient(
            server_url='https://specific-api-url.autotask.net'
        )
        self.assertEqual(
            client.server_url,
            'https://specific-api-url.autotask.net'
        )


class TestAPISettings(TestCase):
    API_URL = 'https://localhost/'

    def setUp(self):
        mk.init_zone_info_connection(return_value={
            'url': self.API_URL,
            'webUrl': self.API_URL,
        })

    def test_default_timeout(self):
        client = api.ContactsAPIClient()
        self.assertEqual(client.timeout, 30.0)

    def test_dynamic_batch_size(self):
        method_name = 'djautotask.utils.DjautotaskSettings.get_settings'
        request_settings = {
            'batch_size': 10,
            'timeout': 10.0,
        }
        _, _patch = mk.create_mock_call(method_name, request_settings)
        client = api.ContactsAPIClient()

        self.assertEqual(client.timeout,
                         request_settings['timeout'])
        _patch.stop()

    def test_retry_attempts(self):
        with self.assertRaises(AutotaskAPIError):
            retry_counter = {'count': 0}
            client = api.ContactsAPIClient()
            client.fetch_resource('http://localhost/some-bad-url',
                                  retry_counter=retry_counter)
            self.assertEqual(retry_counter['count'],
                             client.request_settings['max_attempts'])

    @responses.activate
    def test_no_retry_attempts_in_400_range(self):
        client = api.ContactsAPIClient()
        endpoint = "{}{}".format(client.get_api_url(),
                                 'query?search={"filter": []}')

        tested_status_codes = []
        http_400_range = list(range(400, 499))

        # We do actually want to retry 401, as that indicates the
        # zone information may have changed
        http_400_range.pop(1)

        for status_code in http_400_range:

            retry_counter = {'count': 0}
            try:
                mk.get(endpoint, {}, status=status_code)
                client.fetch_resource(None, retry_counter=retry_counter)
            except AutotaskAPIClientError:
                self.assertEqual(retry_counter['count'], 1)
                tested_status_codes.append(status_code)

        self.assertEqual(tested_status_codes, http_400_range)



# Generated by Django 3.1 on 2020-09-23 09:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0075_taskpredecessor'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='allocationcode',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='servicecall',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='servicecalltask',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='servicecalltaskresource',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='servicecallticket',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='servicecallticketresource',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='task',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='tasknote',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='taskpredecessor',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='tasksecondaryresource',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='ticketnote',
            options={'get_latest_by': 'modified'},
        ),
        migrations.AlterModelOptions(
            name='ticketsecondaryresource',
            options={'get_latest_by': 'modified'},
        ),
    ]



import json
import logging

from braces import views
from django import forms
from django.views.generic import View
from django.http import HttpResponse, HttpResponseBadRequest

from djautotask import sync, models
from djautotask.api import AutotaskAPIError

logger = logging.getLogger(__name__)


class CallBackView(views.CsrfExemptMixin,
                   views.JsonRequestResponseMixin, View):

    CALLBACK_TYPES = {
        'ticket': (
            sync.TicketSynchronizer, models.Ticket
        ),
    }

    def post(self, request, *args, **kwargs):
        """
        Add, OR update entity by fetching it from AT. The only real useful
        information we get from the callback is the id of the ticket. Right
        now Autotask only supports callbacks for Tickets, but it looks like
        support could be added for any other entity at some point. We don't
        receive the ticket object, just its ID, current status, the datetime
        it was originally created, and the user domain it came from, which
        is a lot of useless info. Use the ID to sync the ticket.

        Note that we don't get a callback when a ticket is deleted, so it will
        only get scooped up next sync.
        """

        form = CallBackForm(request.POST)

        if not form.is_valid():
            fields = ', '.join(form.errors.keys())
            msg = 'Received callback with missing parameters: {}.'.format(
                fields)
            logger.warning(msg)
            return HttpResponseBadRequest(json.dumps(form.errors))

        entity_id = form.cleaned_data['id']
        synchronizer = sync.TicketSynchronizer

        try:
            self.handle(entity_id, synchronizer)
        except AutotaskAPIError as e:
            logger.error(
                'API call failed in Ticket ID {} callback: '
                '{}'.format(entity_id, e)
            )

        return HttpResponse(status=204)

    def handle(self, entity_id, synchronizer):
        """
        Do the interesting stuff here, so that it can be overridden in
        a child class if needed.
        """
        synchronizer().fetch_sync_by_id(entity_id)


class CallBackForm(forms.Form):
    id = forms.IntegerField()



API_PAGE_DETAILS = {
    "count": 1,
    "requestCount": 500,
    "prevPageUrl": None,
    "nextPageUrl": None
}

API_EMPTY = {
    "items": [],
    "pageDetails": {
        "count": 0,
        "requestCount": 500,
        "prevPageUrl": None,
        "nextPageUrl": None
    }
}

API_EMPTY_FIELDS = {"fields": []}

API_UDF = {
    "fields": [
        {
            'name': "Test UDF",
            'label': "Test UDF",
            'type': "string",
            'length': 8000,
            'isPickList': True,
            'picklistValues': [
                {
                    'value': "1",
                    'label': "One",
                    'isDefaultValue': False,
                    'sortOrder': 0,
                    'parentValue': None,
                    'isActive': True,
                    'isSystem': False
                }
            ]
        },
    ],
}

API_ACCOUNT_ITEMS = [
    {
        'id': 174,
        'userDefinedFields': [],
        'address1': '26 Tech Valley Drive',
        'address2': 'Suite 2',
        'alternatePhone1': '',
        'alternatePhone2': '',
        'city': 'East Greenbush',
        'createDate': '2012-10-24T05:00:00.000Z',
        'fax': '555-555-6677',
        'lastActivityDate': '2012-06-18T13:15:47.000Z',
        'marketSegmentID': 29683456,
        'companyName': 'Autotask Corporation',
        'companyNumber': "289843",
        'ownerResourceID': 29682885,
        'phone': '555-555-5566',
        'postalCode': 12061,
        'sicCode': '',
        'state': 'NY',
        'stockMarket': '',
        'stockSymbol': '',
        'territoryID': 29683453,
        'companyType': 7,
        'webAddress': 'www.autotask.com',
        'isActive': True,
        'isClientPortalActive': True,
        'isTaskFireActive': False,
        'taxID': '',
        'additionalAddressInformation': '',
        'countryID': 237,
        'billToAddressToUse': 1,
        'billToAttention': '',
        'billToAddress1': '26 Tech Valley Drive',
        'billToAddress2': 'Suite 2',
        'billToCity': 'East Greenbush',
        'billToState': 'NY',
        'billToZipCode': 12061,
        'billToCountryID': 237,
        'billToAdditionalAddressInformation': '',
        'quoteTemplateID': 1,
        'quoteEmailMessageID': 2,
        'invoiceTemplateID': 102,
        'invoiceEmailMessageID': 1,
        'currencyID': 1,
        'createdByResourceID': 29682885,
    }
]
API_ACCOUNT = {
    "items": API_ACCOUNT_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_ACCOUNT_PHYSICAL_LOCATION_ITEMS = [
    {
        'id': 55,
        'name': 'Primary Location',
        'companyID': 174,
        'isActive': True,
        'isPrimary': True
    }
]
API_ACCOUNT_PHYSICAL_LOCATION = {
    "items": API_ACCOUNT_PHYSICAL_LOCATION_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_PROJECT_ITEM = {
    'id': 4,
    'projectName': 'Software Project',
    'companyID': 174,
    'projectType': 5,
    'extProjectNumber': '',
    'projectNumber': 'P20120604.0001',
    'description': '',
    'department': 29683384,
    'contractID': 29684183,
    'createDateTime': '2012-06-18T06:00:00.000Z',
    'creatorResourceID': 4,
    'startDateTime': '2012-06-19T06:00:00.000Z',
    'endDateTime': '2012-10-24T06:00:00.000Z',
    'duration': 59,
    'actualHours': 0.0,
    'actualBilledHours': 0.0,
    'estimatedTime': 164.0,
    'laborEstimatedRevenue': 0.0,
    'laborEstimatedCosts': 0.0,
    'laborEstimatedMarginPercentage': 0.0,
    'projectCostsRevenue': 0.0,
    'projectCostsBudget': 0.0,
    'projectCostEstimatedMarginPercentage': 0.0,
    'changeOrdersRevenue': 0.0,
    'sgda': 0.0,
    'originalEstimatedRevenue': 0.0,
    'estimatedSalesCost': 0.0,
    'status': 1,
    'projectLeadResourceID': 10,
    'completedPercentage': 0,
    'completedDateTime': '2019-09-18T06:00:00.000Z',
    'statusDetail': '',
    'statusDateTime': '2012-06-18T06:00:00.000Z',
    'LineOfBusiness': 6,
    'purchaseOrderNumber': '',
    'businessDivisionSubdivisionID': 6,
    'lastActivityResourceID': 4,
    'lastActivityDateTime': '2012-06-18T02:00:00.000Z',
    'lastActivityPersonType': 1,
    'userDefinedFields': {},
}
API_PROJECT_ITEMS = [API_PROJECT_ITEM]
API_PROJECT_BY_ID = {
    "item": API_PROJECT_ITEM
}
API_PROJECT = {
    "items": API_PROJECT_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_LICENSE_TYPE_FIELD = {
    "fields": [
        {
            "name": "licenseType",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": True,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "1",
                    "label": "Administrator",
                    "isDefaultValue": False,
                    "sortOrder": 1,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                }
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_RESOURCE_ITEMS = [
    {
        'accountingReferenceID': '',
        'isActive': True,
        'dateFormat': 'MM/dd/yyyy',
        'email': '',
        'email2': '',
        'email3': '',
        'emailTypeCode': 'PRIMARY',
        'firstName': 'Autotask',
        'hireDate': '',
        'homePhone': '',
        'initials': '',
        'internalCost': 1.0,
        'lastName': 'Administrator',
        'licenseType': 1,
        'locationID': 90682,
        'middleName': '',
        'mobilePhone': '',
        'numberFormat': 'X,XXX.XX',
        'officeExtension': '',
        'officePhone': '(518) 720-3500',
        'payrollType': 1,
        'resourceType': 'Employee',
        'timeFormat': 'hh:mm a',
        'title': '',
        'userName': 'administrator',
        'userType': 10,
        'id': 10,
    }
]
API_RESOURCE = {
    "items": API_RESOURCE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_SOURCE_FIELD = {
    "fields": [
        {
            "name": "source",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "2",
                    "label": "Insourced",
                    "isDefaultValue": False,
                    "sortOrder": 8,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_ISSUE_TYPE_FIELD = {
    "fields": [
        {
            "name": "issueType",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "2",
                    "label": "Hardware",
                    "isDefaultValue": False,
                    "sortOrder": 1,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_SUB_ISSUE_TYPE_FIELD = {
    "fields": [
        {
            "name": "subIssueType",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    'isActive': True,
                    'isDefaultValue': False,
                    'isSystem': False,
                    'label': 'Rapid Response',
                    'sortOrder': 5,
                    'value': 2,
                    'parentValue': None,
                },
                {
                    'isActive': True,
                    'isDefaultValue': False,
                    'isSystem': False,
                    'label': 'Hardware Request',
                    'sortOrder': 7,
                    'value': 3,
                    'parentValue': None,
                },
            ],
            "picklistParentValueField": "IssueType",
            "isSupportedWebhookField": False
        }
    ]
}

API_TICKET_TYPE_FIELD = {
    "fields": [
        {
            "name": "ticketType",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    'isActive': True,
                    'isDefaultValue': False,
                    'isSystem': False,
                    'label': 'Service Request',
                    'sortOrder': 5,
                    'value': 3,
                    'parentValue': None,
                },
                {
                    'isActive': True,
                    'isDefaultValue': False,
                    'isSystem': False,
                    'label': 'Incident',
                    'sortOrder': 6,
                    'value': 2,
                    'parentValue': None,
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_SERVICE_CALL_STATUS_FIELD = {
    "fields": [
        {
            "name": "status",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "3",
                    "label": "New",
                    "isDefaultValue": False,
                    "sortOrder": 5,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
                {
                    "value": "2",
                    "label": "Complete",
                    "isDefaultValue": False,
                    "sortOrder": 6,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                }
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_QUEUE_FIELD = {
    "fields": [
        {
            "name": "queueID",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "5",
                    "label": "Client Portal",
                    "isDefaultValue": False,
                    "sortOrder": 0,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_PRIORITY_FIELD = {
    "fields": [
        {
            "name": "priority",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "1",
                    "label": "High",
                    "isDefaultValue": False,
                    "sortOrder": 2,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_STATUS_FIELD = {
    "fields": [
        {
            "name": "status",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "9",
                    "label": "Waiting Materials",
                    "isDefaultValue": False,
                    "sortOrder": 6,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
                {
                    "value": "10",
                    "label": "Scheduled",
                    "isDefaultValue": False,
                    "sortOrder": 3,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
                {
                    "value": "11",
                    "label": "Escalate",
                    "isDefaultValue": False,
                    "sortOrder": 5,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
                {
                    "value": "12",
                    "label": "Waiting Vendor",
                    "isDefaultValue": False,
                    "sortOrder": 8,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        },
        {
            "name": "priority",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "1",
                    "label": "High",
                    "isDefaultValue": False,
                    "sortOrder": 2,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        },
        {
            "name": "queueID",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "5",
                    "label": "Client Portal",
                    "isDefaultValue": False,
                    "sortOrder": 0,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        },
        {
            "name": "issueType",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "2",
                    "label": "Hardware",
                    "isDefaultValue": False,
                    "sortOrder": 1,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        },
        {
            "name": "source",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "2",
                    "label": "Insourced",
                    "isDefaultValue": False,
                    "sortOrder": 8,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_TICKET_PICKLIST_FIELD = {
    "fields":
        API_STATUS_FIELD['fields']
        + API_PRIORITY_FIELD['fields']
        + API_QUEUE_FIELD['fields']
        + API_ISSUE_TYPE_FIELD['fields']
        + API_SUB_ISSUE_TYPE_FIELD['fields']
        + API_TICKET_TYPE_FIELD['fields']
        + API_SOURCE_FIELD['fields']
}

API_TICKET_CATEGORY_ITEMS = [
    {
        'id': 5,
        'name': 'Standard (non-editable)',
        'nickname': '',
        'isActive': False,
        'displayColorRGB': 19,
        'isGlobalDefault': False,
        'isApiOnly': False
    },
]
API_TICKET_CATEGORY = {
    "items": API_TICKET_CATEGORY_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_TASK_CATEGORY_FIELD = {
    "fields": [
        {
            "name": "taskCategoryID",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    'value': 100,
                    'label': 'New category',
                    'isDefaultValue': False,
                    'sortOrder': 0,
                    'isActive': True,
                    'isSystem': False,
                    'parentValue': None,
                },
                {
                    'value': 2,
                    'label': 'Standard',
                    'isDefaultValue': True,
                    'sortOrder': 0,
                    'isActive': True,
                    'isSystem': True,
                    'parentValue': None,
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_PROJECT_STATUS_FIELD = {
    "fields": [
        {
            "name": "status",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    'isActive': True,
                    'isDefaultValue': False,
                    'isSystem': True,
                    'label': 'Inactive',
                    'sortOrder': 1,
                    'value': 2,
                    'parentValue': None,
                },
                {
                    'isActive': True,
                    'isDefaultValue': False,
                    'isSystem': True,
                    'label': 'New',
                    'sortOrder': 2,
                    'value': 1,
                    'parentValue': None,
                },
                {
                    'isActive': False,
                    'isDefaultValue': False,
                    'isSystem': True,
                    'label': 'Complete',
                    'sortOrder': 8,
                    'value': 5,
                    'parentValue': None,
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_PROJECT_TYPE_FIELD = {
    "fields": [
        {
            "name": "projectType",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    'isActive': True,
                    'isDefaultValue': False,
                    'isSystem': True,
                    'label': 'Client',
                    'sortOrder': 5,
                    'value': 5,
                    'parentValue': None,
                },
                {
                    'isActive': True,
                    'isDefaultValue': False,
                    'isSystem': True,
                    'label': 'Internal',
                    'sortOrder': 4,
                    'value': 4,
                    'parentValue': None,
                }
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_PROJECT_PICKLIST_FIELD = {
    "fields":
        API_PROJECT_STATUS_FIELD['fields']
        + API_PROJECT_TYPE_FIELD['fields']
}

API_DISPLAY_COLOR_FIELD = {
    "fields": [
        {
            "name": "displayColorRgb",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "19",
                    "label": "#ff6666",
                    "isDefaultValue": False,
                    "sortOrder": 1,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": False
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_ACCOUNT_TYPE_FIELD = {
    "fields": [
        {
            "name": "companyType",
            "dataType": "short",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "3",
                    "label": "Customer",
                    "isDefaultValue": False,
                    "sortOrder": 5,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                }
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_PHASE_ITEMS = [
    {
        'estimatedHours': 23.0,
        'startDate': '2012-08-27T05:00:00.000Z',
        'phaseNumber': 'T20120604.0011',
        'projectID': 4,
        'dueDate': '2012-09-17T05:00:00.000Z',
        'id': 7732,
        'description': 'Unit Testing',
        'createDate': '2012-06-18T17:50:31.000Z',
        'lastActivityDateTime': '2019-11-05T00:28:25.000Z',
        'externalID': None,
        'title': 'Unit Testing',
        'creatorResourceID': 4
    }
]
API_PHASE = {
    "items": API_PHASE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_TASK_ITEM = {
    'id': 101,
    'billingCodeID': 29683415,
    'assignedResourceID': 10,
    'assignedResourceRoleID': 29682834,
    'canClientPortalUserCompleteTask': False,
    'createDateTime': '2018-01-20T12:00:00.000Z',
    'creatorResourceID': 4,
    'completedDateTime': None,
    'departmentID': 29683385,
    'description': 'Review modular code',
    'endDateTime': '2019-09-23T12:00:00.000Z',
    'estimatedHours': 5.0,
    'externalID': None,
    'hoursToBeScheduled': 5.0,
    'isVisibleInClientPortal': True,
    'lastActivityDateTime': '2019-10-06T12:00:00.000Z',
    'phaseID': 7732,
    'priority': 0,
    'priorityLabel': 1,
    'projectID': 4,
    'purchaseOrderNumber': None,
    'remainingHours': 5.0,
    'startDateTime': '2018-01-23T12:00:00.000Z',
    'status': 11,
    'taskIsBillable': False,
    'taskNumber': 'T20120604.0012',
    'taskType': 1,
    'taskCategoryID': 2,
    'title': 'Review modular code',
    'creatorType': 1,
    'lastActivityResourceID': 29683968,
    'lastActivityPersonType': 1,
    'userDefinedFields': {}
}
API_TASK_ITEMS = [API_TASK_ITEM]
API_TASK_BY_ID = {
    "item": API_TASK_ITEM
}
API_TASK = {
    "items": API_TASK_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_TICKET_SECONDARY_RESOURCE_ITEMS = [
    {
        'id': 29684157,
        'ticketID': 100,
        'resourceID': 10,
        'roleID': 8
    },
]
API_TICKET_SECONDARY_RESOURCE = {
    "items": API_TICKET_SECONDARY_RESOURCE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_TASK_SECONDARY_RESOURCE_ITEMS = [
    {
        'id': 29684411,
        'taskID': 101,
        'resourceID': 10,
        'roleID': 8,
    }
]
API_TASK_SECONDARY_RESOURCE = {
    "items": API_TASK_SECONDARY_RESOURCE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_TICKET_NOTE_ITEMS = [
    {
        'id': 45,
        'createDateTime': '2018-01-23T11:00:00.00Z',
        'description': "Note description",
        'creatorResourceID': 10,
        'lastActivityDate': '2018-01-23T11:00:00.00Z',
        'noteType': 2,
        'publish': 1,
        'ticketID': 100,
        'title': "Note Title",
    }
]
API_TICKET_NOTE = {
    "items": API_TICKET_NOTE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_TASK_NOTE_ITEMS = [
    {
        'id': 45,
        'createDateTime': '2018-01-23T11:00:00.00Z',
        'description': "Note description",
        'creatorResourceID': 10,
        'lastActivityDate': '2018-01-23T11:00:00.00Z',
        'noteType': 2,
        'publish': 1,
        'taskID': 101,
        'title': "Note Title",
    }
]
API_TASK_NOTE = {
    "items": API_TASK_NOTE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_NOTE_TYPE_FIELD = {
    "fields": [
        {
            "name": "noteType",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "2",
                    "label": "Task Detail",
                    "isDefaultValue": False,
                    "sortOrder": 1,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                }
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_PROJECT_NOTE_TYPE_FIELD = {
    "fields": [
        {
            "name": "noteType",
            "dataType": "integer",
            "length": 0,
            "isRequired": True,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "5",
                    "label": "Project Notes",
                    "isDefaultValue": False,
                    "sortOrder": 2,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                },
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        },
    ]
}

API_TASK_TYPE_LINK_FIELD = {
    "fields": [
        {
            "name": "timeEntryType",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": True,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "19",
                    "label": "ITServiceRequest",
                    "isDefaultValue": False,
                    "sortOrder": 1,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                }
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_TIME_ENTRY_TICKET_ITEM = {
    'id': 4,
    'ticketID': 100,
    'taskID': None,
    'internalBillingCodeID': 3,
    'timeEntryType': 2,
    'dateWorked': '2018-01-23T00:00:00.00Z',
    'startDateTime': '2018-01-23T10:00:00.00Z',
    'endDateTime': '2018-01-23T12:30:00.00Z',
    'hoursWorked': 1.0000,
    'hoursToBill': 1.0000,
    'offsetHours': 0.0000,
    'summaryNotes': 'Initial triage of issue',
    'internalNotes': 'We will need to get more information',
    'roleID': 8,
    'createDateTime': '2018-01-23T09:50:00.00Z',
    'resourceID': 10,
    'creatorUserID': 10,
    'lastModifiedUserID': 10,
    'lastModifiedDateTime': '2018-01-23T13:00:00.00Z',
    'billingCodeID': 3,
    'contractID': 5,
    'showOnInvoice': True,
    'isNonBillable': False,
    'billingApprovalLevelMostRecent': 0,
}

API_TIME_ENTRY_TASK_ITEM = {
    'id': 5,
    'ticketID': None,
    'taskID': 101,
    'internalBillingCodeID': 3,
    'timeEntryType': 6,
    'dateWorked': '2018-01-23T00:00:00.00Z',
    'startDateTime': '2018-01-23T10:00:00.00Z',
    'endDateTime': '2018-01-23T12:30:00.00Z',
    'hoursWorked': 1.0000,
    'hoursToBill': 1.0000,
    'offsetHours': 0.0000,
    'summaryNotes': 'Entering time for task',
    'internalNotes': 'We will need to get more information',
    'roleID': 8,
    'createDateTime': '2018-01-23T09:50:00.00Z',
    'resourceID': 10,
    'creatorUserID': 10,
    'lastModifiedUserID': 10,
    'lastModifiedDateTime': '2018-01-23T13:00:00.00Z',
    'billingCodeID': 3,
    'contractID': 5,
    'showOnInvoice': True,
    'isNonBillable': False,
    'billingApprovalLevelMostRecent': 0,
}
API_TIME_ENTRY_ITEMS = [API_TIME_ENTRY_TICKET_ITEM, API_TIME_ENTRY_TASK_ITEM]
API_TIME_ENTRY_TICKET = {
    "items": [API_TIME_ENTRY_TICKET_ITEM],
    "pageDetails": API_PAGE_DETAILS
}
API_TIME_ENTRY_TASK = {
    "items": [API_TIME_ENTRY_TASK_ITEM],
    "pageDetails": API_PAGE_DETAILS
}
API_TIME_ENTRY = {
    "items": API_TIME_ENTRY_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_USE_TYPE_FIELD = {
    "fields": [
        {
            "name": "useType",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "2",
                    "label": "General Billing Code",
                    "isDefaultValue": False,
                    "sortOrder": 1,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                }
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_BILLING_CODE_TYPE_FIELD = {
    "fields": [
        {
            "name": "billingCodeType",
            "dataType": "integer",
            "length": 0,
            "isRequired": False,
            "isReadOnly": False,
            "isQueryable": True,
            "isReference": False,
            "referenceEntityType": "",
            "isPickList": True,
            "picklistValues": [
                {
                    "value": "0",
                    "label": "Normal",
                    "isDefaultValue": False,
                    "sortOrder": 1,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                },
                {
                    "value": "1",
                    "label": "System",
                    "isDefaultValue": False,
                    "sortOrder": 2,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                },
                {
                    "value": "2",
                    "label": "Non-Billable",
                    "isDefaultValue": False,
                    "sortOrder": 3,
                    "parentValue": "",
                    "isActive": True,
                    "isSystem": True
                }
            ],
            "picklistParentValueField": "",
            "isSupportedWebhookField": False
        }
    ]
}

API_BILLING_CODE_ITEMS = [
    {
        'id': 2,
        'name': 'Finance',
        'useType': 2,
        'isActive': True,
        'unitCost': 0.0000,
        'unitPrice': 0.0000,
        'externalNumber': 0,
        'isExcludedFromNewContracts': False,
    }
]
API_BILLING_CODE = {
    "items": API_BILLING_CODE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_ROLE_ITEM = {
    'id': 8,
    'name': "IT:Technician I",
    'description': "",
    'isActive': True,
    'hourlyFactor': 1,
    'hourlyRate': 100,
    'roleType': 0,
    'isSystemRole': False,
}
API_ROLE_ITEMS = [API_ROLE_ITEM]
API_ROLE = {
    "items": API_ROLE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_DEPARTMENT_ITEM = {
    'id': 29683384,
    'name': "Finance",
    'description': "Finance Dept",
    'number': "",
}
API_DEPARTMENT_ITEMS = [API_DEPARTMENT_ITEM]
API_DEPARTMENT = {
    "items": API_DEPARTMENT_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_RESOURCE_ROLE_DEPARTMENT_ITEM = {
    "id": 32,
    "departmentID": 29683384,
    "isActive": True,
    "isDefault": True,
    "isDepartmentLead": True,
    "resourceID": 10,
    "roleID": 8,
}
API_RESOURCE_ROLE_DEPARTMENT_ITEMS = [API_RESOURCE_ROLE_DEPARTMENT_ITEM]
API_RESOURCE_ROLE_DEPARTMENT = {
    "items": API_RESOURCE_ROLE_DEPARTMENT_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_RESOURCE_SERVICE_DESK_ROLE_ITEM = {
    "id": 32,
    "isActive": True,
    "isDefault": True,
    "resourceID": 10,
    "roleID": 8,
}
API_RESOURCE_SERVICE_DESK_ROLE_ITEMS = [API_RESOURCE_SERVICE_DESK_ROLE_ITEM]
API_RESOURCE_SERVICE_DESK_ROLE = {
    "items": API_RESOURCE_SERVICE_DESK_ROLE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_CONTRACT_ITEMS = [
    {
        'id': 29684183,
        'userDefinedFields': "",
        'companyID': 174,
        'billingPreference': 2,
        'contractCategory': 15,
        'contractName': "Upstate Document Providers - Hosted SaaS",
        'contractNumber': "2343451",
        'contractPeriodType': "m",
        'contractType': 7,
        'isDefaultContract': True,
        'endDate': '2020-02-23T13:00:00Z',
        'estimatedCost': 0.0000,
        'estimatedHours': 0.0000,
        'estimatedRevenue': 5795.00,
        'setupFee': 995.0000,
        'startDate': '2020-02-23T13:00:00Z',
        'status': 1,
        'timeReportingRequiresStartAndStopTimes': 1,
        'serviceLevelAgreementID': 1,
        'purchaseOrderNumber': "",
        'internalCurrencySetupFee': 995.0000,
    }
]
API_CONTRACT = {
    "items": API_CONTRACT_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_SERVICE_CALL_ITEMS = [
    {
        'id': 2,
        'description': 'Email just in, printer is down.',
        'isComplete': 0,
        'duration': 1.0000,
        'createDateTime': '2020-01-22T13:00:00Z',
        'startDateTime': '2020-02-23T13:00:00Z',
        'endDateTime': '2020-03-23T13:00:00Z',
        'canceledDateTime': None,
        'lastModifiedDateTime': None,
        'companyID': 174,
        'status': 2,
        'creatorResourceID': 10,
        'canceledByResourceID': 10,
    }
]
API_SERVICE_CALL = {
    "items": API_SERVICE_CALL_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_SERVICE_CALL_TICKET_ITEMS = [
    {
        'id': 4,
        'serviceCallID': 2,
        'ticketID': 100,
    }
]
API_SERVICE_CALL_TICKET = {
    "items": API_SERVICE_CALL_TICKET_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_SERVICE_CALL_TASK_ITEMS = [
    {
        'id': 4,
        'serviceCallID': 2,
        'taskID': 101,
    }
]
API_SERVICE_CALL_TASK = {
    "items": API_SERVICE_CALL_TASK_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_SERVICE_CALL_TICKET_RESOURCE_ITEMS = [
    {
        'id': 4,
        'serviceCallTicketID': 4,
        'resourceID': 10,
    }
]
API_SERVICE_CALL_TICKET_RESOURCE = {
    "items": API_SERVICE_CALL_TICKET_RESOURCE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_SERVICE_CALL_TASK_RESOURCE_ITEMS = [
    {
        'id': 4,
        'serviceCallTaskID': 4,
        'resourceID': 10,
    }
]
API_SERVICE_CALL_TASK_RESOURCE = {
    "items": API_SERVICE_CALL_TASK_RESOURCE_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_TASK_PREDECESSOR_ITEMS = [
    {
        'id': 1,
        'lagDays': 0,
        'predecessorTaskID': API_TASK_ITEM['id'],
        'successorTaskID': 7755,
    }
]
API_TASK_PREDECESSOR = {
    "items": API_TASK_PREDECESSOR_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_CONTACT_ITEMS = [
    {
        "id": 29683589,
        "companyID": 174,
        "emailAddress": "brobertson@autotaskdemo.com",
        "emailAddress2": "brobertson2@autotaskdemo.com",
        "emailAddress3": "brobertson3@autotaskdemo.com",
        "firstName": "Mary",
        "isActive": 1,
        "lastActivityDate": "2012-05-25T15:14:29.033Z",
        "lastModifiedDate": "2015-06-10T13:22:29.877Z",
        "lastName": "Smith",
        "phone": "551-555-5513",
        "alternatePhone": "551-555-5522",
        "mobilePhone": "551-555-3502",
    }
]
API_CONTACT = {
    "items": API_CONTACT_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}

API_TICKET_ITEM = {
    'companyID': 174,
    'companyLocationID': 55,
    'billingCodeID': 29683407,
    'assignedResourceID': 10,
    'assignedResourceRoleID': 8,
    'changeInfoField1': '',
    'changeInfoField2': '',
    'changeInfoField3': '',
    'changeInfoField4': '',
    'changeInfoField5': '',
    'completedByResourceID': 29682885,
    'completedDate': '2019-09-05T08:23:00.000Z',
    'contactID': 29683586,
    'contractID': 29684183,
    'createDate': '2019-08-18T01:00:00.000Z',
    'creatorResourceID': 4,
    'creatorType': 1,
    'description': 'Monthy Services Checkup',
    'dueDateTime': '2012-09-07T01:00:00.000Z',
    'estimatedHours': 2.0,
    'externalID': '',
    'firstResponseDateTime': '2012-06-18T01:00:00.000Z',
    'hoursToBeScheduled': 0.0,
    'issueType': 10,
    'lastActivityDate': '2019-09-23T11:00:00.000Z',
    'lastActivityPersonType': 1,
    'lastActivityResourceID': 4,
    'priority': 3,
    'purchaseOrderNumber': '',
    'queueID': 29682833,
    'resolution': '',
    'source': 6,
    'status': 10,
    'subIssueType': 136,
    'ticketCategory': 3,
    'ticketNumber': 'T20120604.0002.005',
    'ticketType': 1,
    'title': 'Monthy Services Checkup*',
    'userDefinedFields': {},
    'id': 100
}
API_TICKET_ITEMS = [API_TICKET_ITEM]
API_TICKET_BY_ID = {
    "item": API_TICKET_ITEM
}
API_TICKET = {
    "items": API_TICKET_ITEMS,
    "pageDetails": API_PAGE_DETAILS
}



# Generated by Django 2.1.14 on 2019-11-27 12:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0027_auto_20191125_1103'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='secondary_resources',
            field=models.ManyToManyField(related_name='secondary_resource_tasks', through='djautotask.TaskSecondaryResource', to='djautotask.Resource'),
        ),
    ]



# Generated by Django 3.1.7 on 2021-08-19 14:02

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0096_merge_20210728_1429'),
    ]

    operations = [
        migrations.AddField(
            model_name='accountphysicallocation',
            name='primary',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='account_physical_location',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djautotask.accountphysicallocation'),
        ),
    ]



# Generated by Django 3.1.1 on 2020-10-01 16:57

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0077_auto_20200924_1512'),
        ('djautotask', '0076_auto_20200925_1326'),
    ]

    operations = [
    ]



# Generated by Django 2.1.14 on 2020-05-05 16:00

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0061_auto_20200416_1241'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='department',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='djautotask.Department'),
        ),
    ]



import pytz

from django.db import models
from django.db.models import Q
from django_extensions.db.models import TimeStampedModel
from django.utils import timezone
from djautotask import api
from model_utils import FieldTracker

from djautotask.api import ProjectsAPIClient, TaskNotesAPIClient, \
    TicketsAPIClient, ServiceCallsAPIClient, TimeEntriesAPIClient, \
    TicketNotesAPIClient

OFFSET_TIMEZONE = 'America/New_York'


class SyncJob(models.Model):
    start_time = models.DateTimeField(null=False)
    end_time = models.DateTimeField(blank=True, null=True)
    entity_name = models.CharField(max_length=100)
    added = models.PositiveIntegerField(null=True)
    updated = models.PositiveIntegerField(null=True)
    skipped = models.PositiveIntegerField(null=True)
    deleted = models.PositiveIntegerField(null=True)
    success = models.BooleanField(null=True)
    message = models.TextField(blank=True, null=True)
    sync_type = models.CharField(max_length=32, default='full')

    def duration(self):
        if self.start_time and self.end_time:
            return self.end_time - self.start_time


class ATUpdateMixin:

    def get_updated_object(self, **kwargs):
        changed_field_keys = kwargs.get('changed_fields')

        updated_object = {}
        if changed_field_keys:
            for field in changed_field_keys:
                updated_object[field] = getattr(self, field)

        return updated_object

    def save(self, *args, **kwargs):
        """
        Save the object.
        If update_at as a kwarg is True, then update Autotask with changes.
        """
        changed_fields = kwargs.pop('changed_fields', None)
        update_at = kwargs.pop('update_at', False)
        impersonation_resource = kwargs.pop('impersonation_resource', None)

        if update_at and changed_fields:
            self.update_at(
                impersonation_resource=impersonation_resource,
                changed_fields=changed_fields
            )

        super().save(**kwargs)


class Ticket(ATUpdateMixin, TimeStampedModel):
    ticket_number = models.CharField(blank=True, null=True, max_length=50)
    completed_date = models.DateTimeField(blank=True, null=True)
    create_date = models.DateTimeField(blank=True, null=True)
    description = models.TextField(blank=True, null=True, max_length=8000)
    due_date_time = models.DateTimeField(null=False)
    estimated_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    last_activity_date = models.DateTimeField(blank=True, null=True)
    title = models.CharField(blank=True, null=True, max_length=255)
    first_response_date_time = models.DateTimeField(blank=True, null=True)
    first_response_due_date_time = models.DateTimeField(blank=True, null=True)
    resolution_plan_date_time = models.DateTimeField(blank=True, null=True)
    resolution_plan_due_date_time = models.DateTimeField(blank=True, null=True)
    resolved_date_time = models.DateTimeField(blank=True, null=True)
    resolved_due_date_time = models.DateTimeField(blank=True, null=True)
    service_level_agreement = models.IntegerField(blank=True, null=True)
    service_level_agreement_has_been_met = models.BooleanField(default=False)
    service_level_agreement_paused_next_event_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    checklist_completed = models.PositiveSmallIntegerField(
        blank=True, null=True)
    checklist_total = models.PositiveSmallIntegerField(blank=True, null=True)

    status = models.ForeignKey(
        'Status', blank=True, null=True, on_delete=models.SET_NULL
    )
    priority = models.ForeignKey(
        'Priority', blank=True, null=True, on_delete=models.SET_NULL
    )
    assigned_resource = models.ForeignKey(
        'Resource', blank=True, null=True, on_delete=models.SET_NULL
    )
    secondary_resources = models.ManyToManyField(
        'Resource', through='TicketSecondaryResource',
        related_name='secondary_resource_tickets'
    )
    queue = models.ForeignKey(
        'Queue', blank=True, null=True, on_delete=models.SET_NULL
    )
    contact = models.ForeignKey(
        'Contact', blank=True, null=True, on_delete=models.SET_NULL
    )
    account = models.ForeignKey(
        'Account', blank=True, null=True, on_delete=models.SET_NULL
    )
    account_physical_location = models.ForeignKey(
        'AccountPhysicalLocation', blank=True, null=True,
        on_delete=models.SET_NULL
    )
    project = models.ForeignKey(
        'Project', blank=True, null=True, on_delete=models.SET_NULL
    )
    category = models.ForeignKey(
        'TicketCategory', blank=True, null=True, on_delete=models.SET_NULL
    )
    source = models.ForeignKey(
        'Source', blank=True, null=True, on_delete=models.SET_NULL
    )
    issue_type = models.ForeignKey(
        'IssueType', blank=True, null=True, on_delete=models.SET_NULL
    )
    sub_issue_type = models.ForeignKey(
        'SubIssueType', blank=True, null=True, on_delete=models.SET_NULL
    )
    type = models.ForeignKey(
        'TicketType', blank=True, null=True, on_delete=models.SET_NULL
    )
    assigned_resource_role = models.ForeignKey(
        'Role', blank=True, null=True, on_delete=models.SET_NULL
    )
    billing_code = models.ForeignKey(
        'BillingCode', null=True, blank=True, on_delete=models.SET_NULL
    )
    contract = models.ForeignKey(
        'Contract', null=True, blank=True, on_delete=models.SET_NULL
    )
    udf = models.JSONField(blank=True, null=True, default=dict)

    AUTOTASK_FIELDS = {
        'title': 'title',
        'description': 'description',
        'queue': 'queueID',
        'estimated_hours': 'estimatedHours',
        'due_date_time': 'dueDateTime',
        'status': 'status',
        'priority': 'priority',
        'category': 'ticketCategory',
        'billing_code': 'billingCodeID',
        'issue_type': 'issueType',
        'sub_issue_type': 'subIssueType',
        'project': 'projectID',
        'assigned_resource': 'assignedResourceID',
        'assigned_resource_role': 'assignedResourceRoleID',
        'account': 'companyID',
        'account_physical_location': 'companyLocationID',
        'contact': 'contactID',
        'contract': 'contractID',
    }

    class Meta:
        verbose_name = 'Ticket'

    def __str__(self):
        return '{}-{}'.format(self.id, self.title)

    def get_account(self):
        return self.account

    def update_at(self, **kwargs):

        api_client = api.TicketsAPIClient(
            impersonation_resource=kwargs.get('impersonation_resource'),
        )
        return api_client.legacy_update(
            self,
            self.get_updated_object(**kwargs)
        )


class AvailablePicklistManager(models.Manager):
    """ Return only active Picklist objects. """

    def get_queryset(self):
        return super().get_queryset().filter(is_active=True)


class Picklist(TimeStampedModel):
    label = models.CharField(blank=True, null=True, max_length=50)
    is_default_value = models.BooleanField(default=False)
    sort_order = models.PositiveIntegerField(blank=True, null=True)
    is_active = models.BooleanField(default=False)
    is_system = models.BooleanField(default=False)

    objects = models.Manager()
    available_objects = AvailablePicklistManager()

    class Meta:
        ordering = ('label',)
        abstract = True

    def __str__(self):
        return self.label if self.label else self.pk


class Status(Picklist):
    # Ticket/task statuses New, Waiting Customer, and Complete are
    # system statuses in Autotask that cannot be edited or deactivated.
    COMPLETE_ID = 5

    class Meta:
        ordering = ('label',)
        verbose_name_plural = 'Statuses'

    def is_complete(self):
        return self.id == Status.COMPLETE_ID


class Priority(Picklist):

    class Meta:
        ordering = ('sort_order',)
        verbose_name_plural = 'Priorities'


class Queue(Picklist):
    label = models.CharField(blank=True, null=True, max_length=100)

    class Meta:
        ordering = ('label',)
        verbose_name_plural = 'Queues'


class ProjectStatus(Picklist):
    COMPLETE_ID = 5

    class Meta:
        ordering = ('label',)
        verbose_name_plural = 'Project statuses'

    def is_complete(self):
        return self.id == ProjectStatus.COMPLETE_ID


class DisplayColor(Picklist):

    class Meta:
        ordering = ('label',)
        verbose_name_plural = 'Display colors'


class ProjectType(Picklist):
    TEMPLATE = 'Template'
    BASELINE = 'Baseline'


class Source(Picklist):
    pass


class IssueType(Picklist):
    pass


class TicketType(Picklist):
    pass


class TaskCategory(Picklist):
    class Meta:
        verbose_name_plural = 'Task categories'


class SubIssueType(Picklist):
    parent_value = models.ForeignKey(
        'IssueType', blank=True, null=True, on_delete=models.CASCADE
    )

    def __str__(self):
        return "{}/{}".format(self.parent_value, self.label) \
            if self.parent_value else self.label


class LicenseType(Picklist):
    # We expect this id's is fixed and immutable.
    impersonation_limited_type = {
        'TEAM_MEMBER': 4,
        'CONTRACTOR': 8,
        'CO_HELP_DESK': 9,
    }
    impersonation_disabled_type = {
        'TIME_ATTENDANCE': 5,
        'DASHBOARD': 6,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_limited_impersonation_targets()

    def has_impersonation(self, target):
        if self.id in self.impersonation_disabled_type.values():
            return False

        if self.id in self.impersonation_limited_type.values():
            return self.check_impersonation(target)

        return True

    def init_limited_impersonation_targets(self):

        if self.id == self.impersonation_limited_type['TEAM_MEMBER']:
            self.limited_impersonation_targets = [
                ProjectsAPIClient,
                TaskNotesAPIClient,
            ]
        elif self.id == \
                self.impersonation_limited_type['CONTRACTOR']:
            self.limited_impersonation_targets = [
                TicketsAPIClient,
                ProjectsAPIClient,
                ServiceCallsAPIClient,
                TimeEntriesAPIClient,
                TicketNotesAPIClient,
                TaskNotesAPIClient,
            ]
        elif self.id == \
                self.impersonation_limited_type['CO_HELP_DESK']:
            self.limited_impersonation_targets = [
                TicketsAPIClient,
                ProjectsAPIClient,
                TaskNotesAPIClient,
            ]
        else:
            self.limited_impersonation_targets = []

    def check_impersonation(self, target):
        for limited_target in self.limited_impersonation_targets:
            if isinstance(target, limited_target):
                return False

        return True


class AccountType(Picklist):
    pass


class NoteType(Picklist):
    SUMMARY = 1
    DETAIL = 2
    NOTES = 3
    # Workflow Rule Note - Task is an Autotask system note type that cannot
    # be edited or deactivated.
    WORKFLOW_RULE_NOTE_ID = 13
    RMM_NOTE = 99


class ProjectNoteType(Picklist):
    NOTES = 5
    EMAIL = 8
    STATUS = 12


class TaskTypeLink(Picklist):
    pass


class UseType(Picklist):
    pass


class BillingCodeType(Picklist):
    NORMAL = 0
    SYSTEM = 1
    NON_BILLABLE = 2


class ServiceCallStatus(Picklist):

    class Meta:
        ordering = ('label',)
        verbose_name_plural = 'Service call statuses'


class TaskType(Picklist):
    pass


class RegularResourceManager(models.Manager):
    API_USER_LICENSE_ID = 7

    def get_queryset(self):
        return super().get_queryset().exclude(
            license_type=self.API_USER_LICENSE_ID)


class Resource(TimeStampedModel):
    user_name = models.CharField(max_length=32)
    email = models.CharField(max_length=250)
    first_name = models.CharField(max_length=50)
    last_name = models.CharField(max_length=50)
    active = models.BooleanField(default=False)
    license_type = models.ForeignKey(
        'LicenseType', null=True, on_delete=models.SET_NULL
    )
    default_service_desk_role = models.ForeignKey(
        'Role', null=True, blank=True, on_delete=models.SET_NULL
    )

    objects = models.Manager()
    regular_objects = RegularResourceManager()

    class Meta:
        ordering = ('first_name', 'last_name')

    def __str__(self):
        return '{} {}'.format(self.first_name, self.last_name)


class TicketCategory(TimeStampedModel):
    # default system id to be used for creating tickets
    STANDARD_ID = 3

    name = models.CharField(max_length=30)
    active = models.BooleanField(default=False)
    display_color = models.ForeignKey(
        'DisplayColor', null=True, on_delete=models.SET_NULL
    )

    class Meta:
        ordering = ('name',)
        verbose_name_plural = 'Ticket categories'

    def __str__(self):
        return self.name


class TicketSecondaryResource(TimeStampedModel):
    resource = models.ForeignKey(
        'Resource', blank=True, null=True, on_delete=models.SET_NULL
    )
    ticket = models.ForeignKey(
        'Ticket', blank=True, null=True, on_delete=models.SET_NULL
    )
    role = models.ForeignKey(
        'Role', null=True, blank=True, on_delete=models.SET_NULL
    )

    AUTOTASK_FIELDS = {
        'resource': 'resourceID',
        'ticket': 'ticketID',
        'role': 'roleID',
    }

    def __str__(self):
        return '{} {}'.format(self.resource, self.ticket)


class Note:
    ALL_USERS = 1
    INTERNAL_USERS = 2
    PUBLISH_CHOICES = ((ALL_USERS, 'All Autotask Users'),
                       (INTERNAL_USERS, 'Internal Users'))
    AUTOTASK_FIELDS = {
        'title': 'title',
        'description': 'description',
        'note_type': 'noteType',
        'publish': 'publish',
        'created_by_contact_id': 'createdByContactID',
    }


class TicketNote(TimeStampedModel, Note):
    title = models.CharField(max_length=250)
    description = models.CharField(max_length=3200)
    create_date_time = models.DateTimeField(blank=True, null=True)
    last_activity_date = models.DateTimeField(blank=True, null=True)

    note_type = models.ForeignKey(
        'NoteType', blank=True, null=True, on_delete=models.SET_NULL)
    creator_resource = models.ForeignKey(
        'Resource', blank=True, null=True, on_delete=models.SET_NULL)
    publish = models.CharField(
        max_length=20,
        blank=True,
        null=True,
        choices=Note.PUBLISH_CHOICES
    )
    ticket = models.ForeignKey(
        'Ticket', blank=True, null=True, on_delete=models.SET_NULL
    )
    created_by_contact = models.ForeignKey(
        'Contact', blank=True, null=True, on_delete=models.SET_NULL
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        super().AUTOTASK_FIELDS.update({
            'ticket': 'ticketID',
        })


class TaskNote(TimeStampedModel, Note):
    title = models.CharField(max_length=250)
    description = models.CharField(max_length=3200)
    create_date_time = models.DateTimeField(blank=True, null=True)
    last_activity_date = models.DateTimeField(blank=True, null=True)

    note_type = models.ForeignKey(
        'NoteType', blank=True, null=True, on_delete=models.SET_NULL)
    creator_resource = models.ForeignKey(
        'Resource', blank=True, null=True, on_delete=models.SET_NULL)
    publish = models.CharField(
        max_length=20,
        blank=True,
        null=True,
        choices=Note.PUBLISH_CHOICES)
    task = models.ForeignKey(
        'Task', blank=True, null=True, on_delete=models.SET_NULL
    )
    created_by_contact = models.ForeignKey(
        'Contact', blank=True, null=True, on_delete=models.SET_NULL
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        super().AUTOTASK_FIELDS.update({
            'task': 'taskID',
        })


class Contact(TimeStampedModel):

    first_name = models.CharField(blank=True, null=True, max_length=200,
                                  db_index=True)
    last_name = models.CharField(blank=True, null=True, max_length=200,
                                 db_index=True)
    email_address = models.CharField(blank=True, null=True, max_length=200)
    email_address2 = models.CharField(blank=True, null=True, max_length=200)
    email_address3 = models.CharField(blank=True, null=True, max_length=200)
    phone = models.CharField(blank=True, null=True, max_length=200)
    alternate_phone = models.CharField(blank=True, null=True, max_length=200)
    mobile_phone = models.CharField(blank=True, null=True, max_length=200)
    extension = models.CharField(blank=True, null=True, max_length=10)
    account = models.ForeignKey(
        'Account', blank=True, null=True, on_delete=models.SET_NULL
    )

    class Meta:
        ordering = ('first_name', 'last_name')

    def __str__(self):
        return '{} {}'.format(self.first_name,
                              self.last_name if self.last_name else '')


class Account(TimeStampedModel):
    MY_ACCOUNT = 0
    name = models.CharField(max_length=100,
                            db_index=True)
    number = models.CharField(max_length=50)
    active = models.BooleanField(default=True)
    last_activity_date = models.DateTimeField(blank=True, null=True)
    phone = models.CharField(blank=True, null=True, max_length=200)
    type = models.ForeignKey(
        'AccountType', blank=True, null=True, on_delete=models.SET_NULL
    )
    parent_account = models.ForeignKey(
        'Account', blank=True, null=True, on_delete=models.SET_NULL
    )
    owner_resource = models.ForeignKey(
        'Resource', blank=True, null=True, on_delete=models.SET_NULL
    )

    class Meta:
        ordering = ('name',)

    def __str__(self):
        return self.name


class AccountPhysicalLocation(models.Model):
    name = models.CharField(max_length=100)
    active = models.BooleanField(default=True)
    primary = models.BooleanField(default=True)
    account = models.ForeignKey(
        'Account', blank=True, null=True, on_delete=models.SET_NULL
    )

    class Meta:
        ordering = ('name',)

    def __str__(self):
        return self.name


class AvailableProjectManager(models.Manager):
    """
    Exclude projects whose type is 'Template' or 'Baseline'. Neither of these
    types provide any use for us, so just exclude them.
    """
    def get_queryset(self):
        return super().get_queryset().exclude(
            Q(type__label=ProjectType.TEMPLATE) |
            Q(type__label=ProjectType.BASELINE)
        )


class Project(ATUpdateMixin, TimeStampedModel):
    name = models.CharField(max_length=100)
    number = models.CharField(null=True, max_length=50)
    description = models.CharField(max_length=2000)
    actual_hours = models.DecimalField(
        null=True, decimal_places=2, max_digits=9)
    completed_date = models.DateField(null=True)
    completed_percentage = models.PositiveSmallIntegerField(default=0)
    duration = models.PositiveIntegerField(default=0)
    start_date = models.DateField(null=True)
    end_date = models.DateField(null=True)
    estimated_time = models.DecimalField(
        null=True, decimal_places=2, max_digits=9
    )
    last_activity_date_time = models.DateTimeField(null=True)
    status_detail = models.CharField(max_length=2000, blank=True, null=True)

    project_lead_resource = models.ForeignKey(
        'Resource', null=True, on_delete=models.SET_NULL
    )
    contact = models.ForeignKey(
        'Contact', blank=True, null=True, on_delete=models.SET_NULL
    )
    account = models.ForeignKey(
        'Account', null=True, on_delete=models.SET_NULL
    )
    status = models.ForeignKey(
        'ProjectStatus', null=True, on_delete=models.SET_NULL
    )
    type = models.ForeignKey(
        'ProjectType', null=True, on_delete=models.SET_NULL
    )
    contract = models.ForeignKey(
        'Contract', null=True, blank=True, on_delete=models.SET_NULL
    )
    department = models.ForeignKey(
        'Department', null=True, blank=True, on_delete=models.SET_NULL
    )
    udf = models.JSONField(blank=True, null=True, default=dict)

    AUTOTASK_FIELDS = {
        'name': 'projectName',
        'status': 'status',
        'description': 'description',
        'start_date': 'startDateTime',
        'end_date': 'endDateTime',
        'type': 'projectType',
        'project_lead_resource': 'projectLeadResourceID',
        'department': 'department',
        'status_detail': 'statusDetail',
    }

    objects = models.Manager()
    available_objects = AvailableProjectManager()

    class Meta:
        ordering = ('name',)

    def __str__(self):
        return self.name

    def update_at(self, **kwargs):
        api_client = api.ProjectsAPIClient(
            impersonation_resource=kwargs.get('impersonation_resource'),
        )
        return api_client.update(
            self,
            self.get_updated_object(**kwargs)
        )


class Phase(TimeStampedModel):
    title = models.CharField(blank=True, null=True, max_length=255)
    description = models.CharField(blank=True, null=True, max_length=8000)
    start_date = models.DateTimeField(blank=True, null=True)
    # due_date is end date in autotask UI
    due_date = models.DateTimeField(blank=True, null=True)
    estimated_hours = models.DecimalField(default=0.0, decimal_places=2,
                                          max_digits=9)
    number = models.CharField(blank=True, null=True, max_length=50)
    scheduled = models.BooleanField(default=False)
    last_activity_date = models.DateTimeField(blank=True, null=True)

    parent_phase = models.ForeignKey(
        'self', null=True, on_delete=models.SET_NULL
    )

    project = models.ForeignKey(
        'Project', null=True, on_delete=models.SET_NULL
    )

    class Meta:
        ordering = ('title',)

    def __str__(self):
        return self.title


class Task(ATUpdateMixin, TimeStampedModel):
    MAX_DESCRIPTION = 8000
    title = models.CharField(blank=True, null=True, max_length=255)
    number = models.CharField(blank=True, null=True, max_length=50)
    description = models.TextField(blank=True, null=True,
                                   max_length=MAX_DESCRIPTION)
    completed_date = models.DateTimeField(blank=True, null=True)
    create_date = models.DateTimeField(blank=True, null=True)
    start_date = models.DateTimeField(blank=True, null=True)
    end_date = models.DateTimeField(blank=True, null=True)
    estimated_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    remaining_hours = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    last_activity_date = models.DateTimeField(blank=True, null=True)

    assigned_resource = models.ForeignKey(
        'Resource', null=True, on_delete=models.SET_NULL
    )
    secondary_resources = models.ManyToManyField(
        'Resource', through='TaskSecondaryResource',
        related_name='secondary_resource_tasks'
    )
    project = models.ForeignKey(
        'Project', on_delete=models.CASCADE
    )
    priority = models.ForeignKey(
        'Priority', null=True, blank=True, on_delete=models.SET_NULL
    )
    status = models.ForeignKey(
        'Status', null=True, on_delete=models.SET_NULL
    )
    phase = models.ForeignKey(
        'Phase', null=True,
        on_delete=models.SET_NULL
    )
    billing_code = models.ForeignKey(
        'BillingCode', null=True, blank=True, on_delete=models.SET_NULL
    )
    assigned_resource_role = models.ForeignKey(
        'Role', blank=True, null=True, on_delete=models.SET_NULL
    )
    department = models.ForeignKey(
        'Department', blank=True, null=True, on_delete=models.SET_NULL
    )
    category = models.ForeignKey(
        'TaskCategory', null=True, blank=True, on_delete=models.SET_NULL
    )
    task_type = models.ForeignKey(
        'TaskType', null=True, blank=True, on_delete=models.SET_NULL
    )
    udf = models.JSONField(blank=True, null=True, default=dict)

    AUTOTASK_FIELDS = {
        'title': 'title',
        'description': 'description',
        'start_date': 'startDateTime',
        'end_date': 'endDateTime',
        'estimated_hours': 'estimatedHours',
        'remaining_hours': 'remainingHours',
        'status': 'status',
        'department': 'departmentID',
        'billing_code': 'billingCodeID',
        'priority': 'priorityLabel',
        'phase': 'phaseID',
        'assigned_resource': 'assignedResourceID',
        'assigned_resource_role': 'assignedResourceRoleID',
        'category': 'taskCategoryID',
    }

    def __str__(self):
        return self.title

    def get_account(self):
        return getattr(self.project, 'account', None)

    def update_at(self, **kwargs):

        api_client = api.TasksAPIClient(
            impersonation_resource=kwargs.get('impersonation_resource'),
        )
        return api_client.legacy_update(
            self,
            self.project,
            self.get_updated_object(**kwargs),
        )


class TaskSecondaryResource(TimeStampedModel):
    resource = models.ForeignKey(
        'Resource', blank=True, null=True, on_delete=models.SET_NULL
    )
    task = models.ForeignKey(
        'Task', blank=True, null=True, on_delete=models.SET_NULL
    )
    role = models.ForeignKey(
        'Role', null=True, blank=True, on_delete=models.SET_NULL
    )

    AUTOTASK_FIELDS = {
        'resource': 'resourceID',
        'task': 'taskID',
        'role': 'roleID',
    }

    def __str__(self):
        return '{} {}'.format(self.resource, self.task)


class TimeEntry(TimeStampedModel):
    date_worked = models.DateTimeField(blank=True, null=True)
    start_date_time = models.DateTimeField(blank=True, null=True)
    end_date_time = models.DateTimeField(blank=True, null=True)
    summary_notes = models.TextField(blank=True, null=True, max_length=8000)
    internal_notes = models.TextField(blank=True, null=True, max_length=8000)
    non_billable = models.BooleanField(default=False)
    hours_worked = models.DecimalField(
        blank=True, null=True, decimal_places=4, max_digits=9)
    hours_to_bill = models.DecimalField(
        blank=True, null=True, decimal_places=4, max_digits=9)
    offset_hours = models.DecimalField(
        blank=True, null=True, decimal_places=4, max_digits=9)

    resource = models.ForeignKey(
        'Resource', blank=True, null=True, on_delete=models.CASCADE)
    ticket = models.ForeignKey(
        'Ticket', blank=True, null=True, on_delete=models.CASCADE)
    task = models.ForeignKey(
        'Task', blank=True, null=True, on_delete=models.CASCADE)
    type = models.ForeignKey(
        'TaskTypeLink', blank=True, null=True, on_delete=models.SET_NULL)
    billing_code = models.ForeignKey(
        'BillingCode', blank=True, null=True, on_delete=models.SET_NULL)
    role = models.ForeignKey(
        'Role', blank=True, null=True, on_delete=models.SET_NULL
    )
    contract = models.ForeignKey(
        'Contract', blank=True, null=True, on_delete=models.SET_NULL
    )

    AUTOTASK_FIELDS = {
        'ticket': 'ticketID',
        'task': 'taskID',
        'date_worked': 'dateWorked',
        'start_date_time': 'startDateTime',
        'end_date_time': 'endDateTime',
        'summary_notes': 'summaryNotes',
        'internal_notes': 'internalNotes',
        'hours_worked': 'hoursWorked',
        'offset_hours': 'offsetHours',
        'role': 'roleID',
        'resource': 'resourceID',
        'billing_code': 'billingCodeID',
        'contract': 'contractID',
        'show_on_invoice': 'showOnInvoice',
    }

    class Meta:
        verbose_name_plural = 'Time entries'
        ordering = ('-start_date_time', '-date_worked', '-id')

    def __str__(self):
        return str(self.id) or ''

    def get_entered_time(self):
        """
        In Autotask, tickets are required to have start and end times.
        Start and end times for project tasks can be optional.
        In the case that a task has no start or end time, use the
        date_worked field.
        """
        if self.end_date_time:
            entered_time = self.end_date_time
        else:
            # Autotask gives us date_worked as a datetime, even though the
            # time is always set to EST midnight (00:00:00).
            # TODO timezone.pytz does not exist anymore
            est_offset = timezone.localtime(
                timezone=pytz.timezone(OFFSET_TIMEZONE)).utcoffset()
            local_offset = timezone.localtime().utcoffset()

            # We want to end up with a UTC datetime that is midnight in the
            # local timezone.
            date_worked = self.date_worked + est_offset
            entered_time = date_worked - local_offset

        return entered_time


class BillingCode(TimeStampedModel):
    # BillingCodes with use type General Allocation Code (with ID = 1)
    # are for setting a ticket's work type in the UI. See API docs for details.
    # https://ww2.autotask.net/help/DeveloperHelp/Content/APIs/REST/Entities/BillingCodesEntity.htm # noqa
    GENERAL_BILLING_CODE_ID = 1
    name = models.CharField(blank=True, null=True, max_length=200)
    description = models.CharField(blank=True, null=True, max_length=500)
    active = models.BooleanField(default=False)

    use_type = models.ForeignKey(
        'UseType', blank=True, null=True, on_delete=models.SET_NULL
    )
    billing_code_type = models.ForeignKey(
        'BillingCodeType', blank=True, null=True, on_delete=models.SET_NULL
    )

    def __str__(self):
        return self.name if self.name else self.pk


class Role(models.Model):
    name = models.CharField(max_length=200)
    active = models.BooleanField(default=True)
    description = models.CharField(blank=True, null=True, max_length=200)
    hourly_factor = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    hourly_rate = models.DecimalField(
        blank=True, null=True, decimal_places=2, max_digits=9)
    role_type = models.PositiveIntegerField(blank=True, null=True)
    system_role = models.BooleanField(default=False)

    def __str__(self):
        return self.name


class Department(models.Model):
    name = models.CharField(max_length=100)
    description = models.TextField(blank=True, null=True, max_length=1000)
    number = models.CharField(blank=True, null=True, max_length=50)

    def __str__(self):
        return self.name


class ResourceRoleDepartment(models.Model):
    id = models.BigAutoField(primary_key=True, verbose_name='ID')
    active = models.BooleanField(default=True)
    default = models.BooleanField(default=False)
    department_lead = models.BooleanField(default=False)

    def __str__(self):
        return "{} - {} - {}".format(
            self.resource, self.role, self.department)

    department = models.ForeignKey(
        'Department', on_delete=models.CASCADE)
    role = models.ForeignKey(
        'Role', on_delete=models.CASCADE)
    resource = models.ForeignKey(
        'Resource', on_delete=models.CASCADE)


class ResourceServiceDeskRole(models.Model):
    active = models.BooleanField(default=True)
    default = models.BooleanField(default=False)

    def __str__(self):
        return "{} - {}".format(
            self.resource, self.role)

    role = models.ForeignKey(
        'Role', on_delete=models.CASCADE)
    resource = models.ForeignKey(
        'Resource', on_delete=models.CASCADE)


class Contract(models.Model):
    INACTIVE = 0
    ACTIVE = 1
    STATUS_CHOICES = (
        (INACTIVE, 'Inactive'),
        (ACTIVE, 'Active')
    )
    name = models.CharField(max_length=250, db_index=True)
    number = models.CharField(blank=True, null=True, max_length=50)
    status = models.CharField(
        max_length=20, blank=True, null=True, choices=STATUS_CHOICES)

    account = models.ForeignKey(
        'Account', blank=True, null=True, on_delete=models.SET_NULL
    )

    class Meta:
        ordering = ('name',)

    def __str__(self):
        return self.name


class ServiceCall(TimeStampedModel):
    description = models.TextField(blank=True, null=True, max_length=2000)
    duration = models.DecimalField(
        blank=True, null=True, decimal_places=4, max_digits=9)
    complete = models.BooleanField(default=False)
    create_date_time = models.DateTimeField(blank=True, null=True)
    start_date_time = models.DateTimeField()
    end_date_time = models.DateTimeField()
    canceled_date_time = models.DateTimeField(blank=True, null=True)
    last_modified_date_time = models.DateTimeField(blank=True, null=True)

    account = models.ForeignKey(
        'Account', blank=True, null=True, on_delete=models.SET_NULL)
    location = models.ForeignKey(
        'AccountPhysicalLocation',
        blank=True,
        null=True,
        on_delete=models.SET_NULL
    )
    status = models.ForeignKey(
        'ServiceCallStatus', on_delete=models.CASCADE)
    creator_resource = models.ForeignKey(
        'Resource',
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name='created_service_calls'
    )
    canceled_by_resource = models.ForeignKey(
        'Resource',
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name='cancelled_service_calls'
    )

    tickets = models.ManyToManyField(
        'Ticket', through='ServiceCallTicket',
        related_name='ticket_service_calls'
    )
    tasks = models.ManyToManyField(
        'Task', through='ServiceCallTask',
        related_name='task_service_calls'
    )

    AUTOTASK_FIELDS = {
        'description': 'description',
        'duration': 'duration',
        'complete': 'isComplete',
        'create_date_time': 'createDateTime',
        'start_date_time': 'startDateTime',
        'end_date_time': 'endDateTime',
        'canceled_date_time': 'canceledDateTime',
        'last_modified_date_time': 'lastModifiedDateTime',
        'account': 'companyID',
        'location': 'companyLocationID',
        'status': 'status',
        'creator_resource': 'creatorResourceID',
        'canceled_by_resource': 'canceledByResourceID',
    }

    def __str__(self):
        return str(self.id)


class ServiceCallTicket(TimeStampedModel):
    service_call = models.ForeignKey(
        'ServiceCall',
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    ticket = models.ForeignKey('Ticket', on_delete=models.CASCADE)
    resources = models.ManyToManyField(
        'Resource', through='ServiceCallTicketResource',
        related_name='resource_service_call_ticket'
    )

    AUTOTASK_FIELDS = {
        'service_call': 'serviceCallID',
        'ticket': 'ticketID',
    }

    def __str__(self):
        return str(self.id)


class ServiceCallTask(TimeStampedModel):
    service_call = models.ForeignKey(
        'ServiceCall',
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )
    task = models.ForeignKey('Task', on_delete=models.CASCADE)
    resources = models.ManyToManyField(
        'Resource', through='ServiceCallTaskResource',
        related_name='resource_service_call_task'
    )

    AUTOTASK_FIELDS = {
        'service_call': 'serviceCallID',
        'task': 'taskID',
    }

    def __str__(self):
        return str(self.id)


class ServiceCallTicketResource(TimeStampedModel):
    service_call_ticket = models.ForeignKey(
        'ServiceCallTicket', on_delete=models.CASCADE)
    resource = models.ForeignKey('Resource', on_delete=models.CASCADE)

    AUTOTASK_FIELDS = {
        'resource': 'resourceID',
        'service_call_ticket': 'serviceCallTicketID',
    }

    def __str__(self):
        return str(self.id)


class ServiceCallTaskResource(TimeStampedModel):
    service_call_task = models.ForeignKey(
        'ServiceCallTask', on_delete=models.CASCADE)
    resource = models.ForeignKey('Resource', on_delete=models.CASCADE)

    AUTOTASK_FIELDS = {
        'resource': 'resourceID',
        'service_call_task': 'serviceCallTaskID',
    }

    def __str__(self):
        return str(self.id)


class TaskPredecessor(TimeStampedModel):
    lag_days = models.IntegerField(blank=True, null=True)
    predecessor_task = models.ForeignKey('Task', blank=True, null=True,
                                         related_name='predecessor_task_set',
                                         on_delete=models.CASCADE)
    successor_task = models.ForeignKey('Task', blank=True, null=True,
                                       related_name='successor_task_set',
                                       on_delete=models.CASCADE)

    class Meta:
        ordering = ('predecessor_task__title', )

    def __str__(self):
        return str(self.id)


class BaseUDF(TimeStampedModel):
    name = models.CharField(max_length=50, unique=True)
    label = models.CharField(max_length=50, blank=True, null=True)
    type = models.CharField(max_length=50, blank=True, null=True)
    is_picklist = models.BooleanField(default=False)
    picklist = models.JSONField(default=dict)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name


class TicketUDF(BaseUDF):
    pass


class TaskUDF(BaseUDF):
    pass


class ProjectUDF(BaseUDF):
    pass


class TicketTracker(Ticket):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_ticket'


class StatusTracker(Status):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_status'


class PriorityTracker(Priority):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_priority'


class QueueTracker(Queue):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_queue'


class ProjectStatusTracker(ProjectStatus):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_projectstatus'


class DisplayColorTracker(DisplayColor):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_displaycolor'


class ProjectTypeTracker(ProjectType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_projecttype'


class ProjectNoteTypeTracker(ProjectNoteType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_projectnotetype'


class SourceTracker(Source):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_source'


class IssueTypeTracker(IssueType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_issuetype'


class TicketTypeTracker(TicketType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_tickettype'


class TaskCategoryTracker(TaskCategory):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_taskcategory'


class SubIssueTypeTracker(SubIssueType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_subissuetype'


class LicenseTypeTracker(LicenseType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_licensetype'


class AccountTypeTracker(AccountType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_accounttype'


class NoteTypeTracker(NoteType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_notetype'


class TaskTypeLinkTracker(TaskTypeLink):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_tasktypelink'


class UseTypeTracker(UseType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_usetype'


class BillingCodeTypeTracker(BillingCodeType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_billingcodetype'


class ServiceCallStatusTracker(ServiceCallStatus):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_servicecallstatus'


class ResourceTracker(Resource):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_resource'


class TicketCategoryTracker(TicketCategory):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_ticketcategory'


class TicketSecondaryResourceTracker(TicketSecondaryResource):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_ticketsecondaryresource'


class TicketNoteTracker(TicketNote):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_ticketnote'


class TaskNoteTracker(TaskNote):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_tasknote'


class ContactTracker(Contact):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_contact'


class AccountTracker(Account):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_account'


class AccountPhysicalLocationTracker(AccountPhysicalLocation):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_accountphysicallocation'


class ProjectTracker(Project):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_project'


class PhaseTracker(Phase):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_phase'


class TaskTracker(Task):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_task'


class TaskSecondaryResourceTracker(TaskSecondaryResource):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_tasksecondaryresource'


class TimeEntryTracker(TimeEntry):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_timeentry'


class BillingCodeTracker(BillingCode):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_billingcode'


class RoleTracker(Role):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_role'


class DepartmentTracker(Department):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_department'


class ResourceRoleDepartmentTracker(ResourceRoleDepartment):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_resourceroledepartment'


class ResourceServiceDeskRoleTracker(ResourceServiceDeskRole):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_resourceservicedeskrole'


class ContractTracker(Contract):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_contract'


class ServiceCallTracker(ServiceCall):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_servicecall'


class ServiceCallTicketTracker(ServiceCallTicket):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_servicecallticket'


class ServiceCallTaskTracker(ServiceCallTask):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_servicecalltask'


class ServiceCallTicketResourceTracker(ServiceCallTicketResource):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_servicecallticketresource'


class ServiceCallTaskResourceTracker(ServiceCallTaskResource):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_servicecalltaskresource'


class TaskPredecessorTracker(TaskPredecessor):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_taskpredecessor'


class TicketUDFTracker(TicketUDF):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_ticketudf'


class TaskUDFTracker(TaskUDF):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_taskudf'


class ProjectUDFTracker(ProjectUDF):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_projectudf'


class TaskTypeTracker(TaskType):
    tracker = FieldTracker()

    class Meta:
        proxy = True
        db_table = 'djautotask_tasktype'



# Generated by Django 3.1.7 on 2021-07-27 09:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djautotask', '0093_auto_20210629_1204'),
    ]

    operations = [
        migrations.AddField(
            model_name='ticket',
            name='checklist_completed',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='ticket',
            name='checklist_total',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
    ]



from djautotask.tests import mocks, fixtures

from djautotask import models
from djautotask import sync


def generate_api_objects(fixture_objects):
    """
    Generate multiple objects based on the given fixtures.
    """
    if fixture_objects is None:
        raise Exception("fixture_objects is empty.")

    fixture_result = {
        "items": [],
        "pageDetails": fixtures.API_PAGE_DETAILS
    }
    fixture_result["pageDetails"]["count"] = 0

    for fixture in fixture_objects:
        fixture_result["items"] += fixture["items"]
        fixture_result["pageDetails"]["count"] += len(fixture["items"])

    return fixture_result


def init_ticket_udfs():
    models.TicketUDF.objects.all().delete()
    mocks.service_api_get_ticket_udf_call(fixtures.API_UDF)
    synchronizer = sync.TicketUDFSynchronizer()
    return synchronizer.sync()


def init_task_udfs():
    models.TaskUDF.objects.all().delete()
    mocks.service_api_get_task_udf_call(fixtures.API_UDF)
    synchronizer = sync.TaskUDFSynchronizer()
    return synchronizer.sync()


def init_project_udfs():
    models.ProjectUDF.objects.all().delete()
    mocks.service_api_get_project_udf_call(fixtures.API_UDF)
    synchronizer = sync.ProjectUDFSynchronizer()
    return synchronizer.sync()


def init_contacts():
    models.Contact.objects.all().delete()
    mocks.service_api_get_contacts_call(fixtures.API_CONTACT)
    synchronizer = sync.ContactSynchronizer()
    return synchronizer.sync()


def init_project_statuses():
    models.ProjectStatus.objects.all().delete()
    mocks.service_api_get_project_picklist_call(
        fixtures.API_PROJECT_STATUS_FIELD)
    synchronizer = sync.ProjectStatusSynchronizer()
    return synchronizer.sync()


def init_project_types():
    models.ProjectType.objects.all().delete()
    mocks.service_api_get_project_picklist_call(
        fixtures.API_PROJECT_TYPE_FIELD)
    synchronizer = sync.ProjectTypeSynchronizer()
    return synchronizer.sync()


def init_display_colors():
    models.DisplayColor.objects.all().delete()
    mocks.service_api_get_ticket_category_picklist_call(
        fixtures.API_DISPLAY_COLOR_FIELD)
    synchronizer = sync.DisplayColorSynchronizer()
    return synchronizer.sync()


def init_issue_types():
    models.IssueType.objects.all().delete()
    mocks.service_api_get_ticket_picklist_call(fixtures.API_ISSUE_TYPE_FIELD)
    synchronizer = sync.IssueTypeSynchronizer()
    return synchronizer.sync()


def init_sub_issue_types():
    models.SubIssueType.objects.all().delete()
    mocks.service_api_get_ticket_picklist_call(
        fixtures.API_SUB_ISSUE_TYPE_FIELD)
    synchronizer = sync.SubIssueTypeSynchronizer()
    return synchronizer.sync()


def init_ticket_types():
    models.TicketType.objects.all().delete()
    mocks.service_api_get_ticket_picklist_call(
        fixtures.API_TICKET_TYPE_FIELD)
    synchronizer = sync.TicketTypeSynchronizer()
    return synchronizer.sync()


def init_statuses():
    models.Status.objects.all().delete()
    mocks.service_api_get_ticket_picklist_call(fixtures.API_STATUS_FIELD)
    synchronizer = sync.StatusSynchronizer()
    return synchronizer.sync()


def init_priorities():
    models.Priority.objects.all().delete()
    mocks.service_api_get_ticket_picklist_call(fixtures.API_PRIORITY_FIELD)
    synchronizer = sync.PrioritySynchronizer()
    return synchronizer.sync()


def init_queues():
    models.Queue.objects.all().delete()
    mocks.service_api_get_ticket_picklist_call(fixtures.API_QUEUE_FIELD)
    synchronizer = sync.QueueSynchronizer()
    return synchronizer.sync()


def init_sources():
    models.Source.objects.all().delete()
    mocks.service_api_get_ticket_picklist_call(fixtures.API_SOURCE_FIELD)
    synchronizer = sync.SourceSynchronizer()
    return synchronizer.sync()


def init_license_types():
    models.LicenseType.objects.all().delete()
    mocks.service_api_get_license_types_call(fixtures.API_LICENSE_TYPE_FIELD)
    synchronizer = sync.LicenseTypeSynchronizer()
    return synchronizer.sync()


def init_use_types():
    models.UseType.objects.all().delete()
    mocks.service_api_get_use_types_call(fixtures.API_USE_TYPE_FIELD)
    synchronizer = sync.UseTypeSynchronizer()
    return synchronizer.sync()


def init_task_categories():
    models.TaskCategory.objects.all().delete()
    mocks.service_api_get_task_picklist_call(fixtures.API_TASK_CATEGORY_FIELD)
    synchronizer = sync.TaskCategorySynchronizer()
    return synchronizer.sync()


def init_task_type_links():
    models.TaskTypeLink.objects.all().delete()
    mocks.service_api_get_task_type_links_call(
        fixtures.API_TASK_TYPE_LINK_FIELD
    )
    synchronizer = sync.TaskTypeLinkSynchronizer()
    return synchronizer.sync()


def init_account_types():
    models.AccountType.objects.all().delete()
    mocks.service_api_get_account_types_call(fixtures.API_ACCOUNT_TYPE_FIELD)
    synchronizer = sync.AccountTypeSynchronizer()
    return synchronizer.sync()


def init_service_call_statuses():
    models.ServiceCallStatus.objects.all().delete()
    mocks.service_api_get_service_call_statuses_call(
        fixtures.API_SERVICE_CALL_STATUS_FIELD)
    synchronizer = sync.ServiceCallStatusSynchronizer()
    return synchronizer.sync()


def init_ticket_categories():
    models.TicketCategory.objects.all().delete()
    mocks.service_api_get_ticket_categories_call(fixtures.API_TICKET_CATEGORY)
    synchronizer = sync.TicketCategorySynchronizer()
    return synchronizer.sync()


def init_task_predecessors():
    models.TaskPredecessor.objects.all().delete()
    mocks.service_api_get_task_predecessors_call(fixtures.API_TASK_PREDECESSOR)
    synchronizer = sync.TaskPredecessorSynchronizer()
    return synchronizer.sync()


def init_tickets():
    models.Ticket.objects.all().delete()
    mocks.service_api_get_tickets_call(fixtures.API_TICKET)
    synchronizer = sync.TicketSynchronizer()
    return synchronizer.sync()


def init_resources():
    models.Resource.objects.all().delete()
    mocks.service_api_get_resources_call(fixtures.API_RESOURCE)
    synchronizer = sync.ResourceSynchronizer()
    return synchronizer.sync()


def init_ticket_secondary_resources():
    models.TicketSecondaryResource.objects.all().delete()
    mocks.service_api_get_ticket_secondary_resources_call(
        fixtures.API_TICKET_SECONDARY_RESOURCE)
    synchronizer = sync.TicketSecondaryResourceSynchronizer()
    return synchronizer.sync()


def init_task_secondary_resources():
    models.TaskSecondaryResource.objects.all().delete()
    mocks.service_api_get_task_secondary_resources_call(
        fixtures.API_TASK_SECONDARY_RESOURCE)
    synchronizer = sync.TaskSecondaryResourceSynchronizer()
    return synchronizer.sync()


def init_accounts():
    models.Account.objects.all().delete()
    mocks.service_api_get_accounts_call(fixtures.API_ACCOUNT)
    synchronizer = sync.AccountSynchronizer()
    return synchronizer.sync()


def init_account_physical_locations():
    models.AccountPhysicalLocation.objects.all().delete()
    mocks.service_api_get_account_physical_locations_call(
        fixtures.API_ACCOUNT_PHYSICAL_LOCATION
    )
    synchronizer = sync.AccountPhysicalLocationSynchronizer()
    return synchronizer.sync()


def init_projects():
    models.Project.objects.all().delete()
    mocks.service_api_get_projects_call(fixtures.API_PROJECT)
    synchronizer = sync.ProjectSynchronizer()
    return synchronizer.sync()


def init_phases():
    models.Phase.objects.all().delete()
    mocks.service_api_get_phases_call(fixtures.API_PHASE)
    synchronizer = sync.PhaseSynchronizer()
    return synchronizer.sync()


def init_tasks():
    models.Task.objects.all().delete()
    mocks.service_api_get_tasks_call(fixtures.API_TASK)
    synchronizer = sync.TaskSynchronizer()
    return synchronizer.sync()


def init_ticket_notes():
    mocks.create_mock_call(
        'djautotask.sync.TicketNoteSynchronizer.create', None)

    models.TicketNote.objects.all().delete()
    mocks.service_api_get_ticket_notes_call(fixtures.API_TICKET_NOTE)
    synchronizer = sync.TicketNoteSynchronizer()
    return synchronizer.sync()


def init_task_notes():
    mocks.create_mock_call(
        'djautotask.sync.TaskNoteSynchronizer.create', None)

    models.TaskNote.objects.all().delete()
    mocks.service_api_get_task_notes_call(fixtures.API_TASK_NOTE)
    synchronizer = sync.TaskNoteSynchronizer()
    return synchronizer.sync()


def init_note_types():
    models.NoteType.objects.all().delete()
    mocks.service_api_get_note_types_call(fixtures.API_NOTE_TYPE_FIELD)
    synchronizer = sync.NoteTypeSynchronizer()
    return synchronizer.sync()


def init_project_note_types():
    models.ProjectNoteType.objects.all().delete()
    mocks.service_api_get_project_note_types_call(
        fixtures.API_PROJECT_NOTE_TYPE_FIELD)
    synchronizer = sync.ProjectNoteTypeSynchronizer()
    return synchronizer.sync()


def init_time_entries():
    models.TimeEntry.objects.all().delete()
    mocks.service_api_get_time_entries_call(fixtures.API_TIME_ENTRY)
    synchronizer = sync.TimeEntrySynchronizer()
    return synchronizer.sync()


def init_billing_codes():
    models.BillingCode.objects.all().delete()
    mocks.service_api_get_billing_codes_call(fixtures.API_BILLING_CODE)
    synchronizer = sync.BillingCodeSynchronizer()
    return synchronizer.sync()


def init_roles():
    models.Role.objects.all().delete()
    mocks.service_api_get_roles_call(fixtures.API_ROLE)
    synchronizer = sync.RoleSynchronizer()
    return synchronizer.sync()


def init_departments():
    models.Department.objects.all().delete()
    mocks.service_api_get_departments_call(fixtures.API_DEPARTMENT)
    synchronizer = sync.DepartmentSynchronizer()
    return synchronizer.sync()


def init_resource_service_desk_roles():
    models.ResourceServiceDeskRole.objects.all().delete()
    mocks.service_api_get_resource_service_desk_roles_call(
        fixtures.API_RESOURCE_SERVICE_DESK_ROLE)
    synchronizer = sync.ResourceServiceDeskRoleSynchronizer()
    return synchronizer.sync()


def init_resource_role_departments():
    models.ResourceRoleDepartment.objects.all().delete()
    mocks.service_api_get_resource_role_departments_call(
        fixtures.API_RESOURCE_ROLE_DEPARTMENT)
    synchronizer = sync.ResourceRoleDepartmentSynchronizer()
    return synchronizer.sync()


def init_contracts():
    models.Contract.objects.all().delete()
    mocks.service_api_get_contracts_call(fixtures.API_CONTRACT)
    synchronizer = sync.ContractSynchronizer()
    return synchronizer.sync()


def init_service_calls():
    models.ServiceCall.objects.all().delete()
    mocks.service_api_get_service_calls_call(fixtures.API_SERVICE_CALL)
    synchronizer = sync.ServiceCallSynchronizer()
    return synchronizer.sync()


def init_service_call_tickets():
    models.ServiceCallTicket.objects.all().delete()
    mocks.service_api_get_service_call_tickets_call(
        fixtures.API_SERVICE_CALL_TICKET)
    synchronizer = sync.ServiceCallTicketSynchronizer()
    return synchronizer.sync()


def init_service_call_tasks():
    models.ServiceCallTask.objects.all().delete()
    mocks.service_api_get_service_call_tasks_call(
        fixtures.API_SERVICE_CALL_TASK)
    synchronizer = sync.ServiceCallTaskSynchronizer()
    return synchronizer.sync()


def init_service_call_ticket_resources():
    models.ServiceCallTicketResource.objects.all().delete()
    mocks.service_api_get_service_call_ticket_resources_call(
        fixtures.API_SERVICE_CALL_TICKET_RESOURCE)
    synchronizer = sync.ServiceCallTicketResourceSynchronizer()
    return synchronizer.sync()


def init_service_call_task_resources():
    models.ServiceCallTaskResource.objects.all().delete()
    mocks.service_api_get_service_call_task_resources_call(
        fixtures.API_SERVICE_CALL_TASK_RESOURCE)
    synchronizer = sync.ServiceCallTaskResourceSynchronizer()
    return synchronizer.sync()
